#
# code-growth-fishing.R, 29 Sep 24
#
# Data from:
#
# A statistical examination of the properties and evolution of libre software
# Israel Herraiz Tabernero
# Studying the laws of software evolution in a long-lived FLOSS project
# Jesus M. Gonzalez-Barahona and Gregorio Robles and Israel Herraiz and Felipe Ortega

library("colorspace")
library("plyr")


par(bty="l")
par(las=1)
par(pch="+")
pal_col=rainbow(3)

# Simple model of program growth
# LOC at time n+1 given by the recurrence relation
# M_{n+1}=a*M_n+b
# M_n=(M_0-b/(1-a))+b/(1-a))
LOC=function(M0, a, b, n=1:50) (M0-b/(1-a))*a^n+b/(1-a)

plot(LOC(0, 0.9, 15), type="l", col=pal_col[1],
	xaxs="i",
	xlim=c(0, 50),
	xlab="n", ylab="LOC")
lines(LOC(0, 0.9, 10), col=pal_col[2])
lines(LOC(0, 0.8, 15), col=pal_col[3])

legend("bottomright", legend=c("a=0.9, b=15", "a=0.9, b=10", "a=0.8, b=15"),
                        fill=pal_col, border="white", bty="n", cex=1.2)

dev.copy(dev=png, file="code-growth-S1.png")
dev.off()

freebsd=read.csv("Herraiz-freebsd.txt.xz", as.is=TRUE)

freebsd$ksloc=freebsd$sloc/1e3
freebsd$date=as.Date(freebsd$date, format="%Y-%m-%d")
start_date=freebsd$date[1]
freebsd$Number_days=as.integer(difftime(freebsd$date,
                                         start_date,
                                         units="days"))
# Order by days since first release

plot(freebsd$Number_days, freebsd$ksloc, col=pal_col[2],
        xaxs="i", yaxs="i",
        xlab="Elapsed days", ylab="Lines of code (thousands)")

lin_mod=glm(ksloc ~ Number_days, data=freebsd)

lin_pred=predict(lin_mod)
lines(lin_pred, col=pal_col[3])

# Fit simple model
ad_mod=nls(ksloc ~ b*(1-a^Number_days)/(1-a)+a^Number_days*26, data=freebsd,
			start=list(a=0.999, b=0.6))
summary(ad_mod)

ad_coef=coef(ad_mod)

# Plot fit
Number_days=0:5000
ad_ksloc = ad_coef[2]*(1-ad_coef[1]^Number_days)/(1-ad_coef[1])+ad_coef[1]^Number_days*26
lines(ad_ksloc, col=pal_col[1])

legend("bottomright", legend=c("Add/delete model", "Data", "Fitted regression line"),
                        fill=pal_col, border="white", bty="n", cex=1.2)

dev.copy(dev=png, file="code-growth-FreeBSD.png")
dev.off()

# asymptotic max lines
ad_coef[2]/(1-ad_coef[1])

week_freebsd=subset(freebsd, Number_days %% 7 == 0)
week_freebsd$Number_weeks=week_freebsd$Number_days %/% 7
wad_mod=nls(ksloc ~ b*(1-a^Number_weeks)/(1-a)+a^Number_weeks*26, data=week_freebsd,
			start=list(a=0.999, b=4.2))
summary(wad_mod)

ad_coef=coef(ad_mod)

# asymptotic max lines
ad_coef[2]/(1-ad_coef[1])


# P_{n+1}=c*P_n+d
# M_{n+1}=a*M_n+b*P_n
# M_{n+1}=a*M_n+b*((P_0-d/(1-c))*c^(n-lag)+d/(1-c))
#
# l1=b/(1-a)*d/(1-c))
# l2=b*(P_0*(1-c)-d)/(c*(1-c))
# M_n=(M_0-l2-l1)*a^n+l2*c^(n-1)+l1

dev_loc=function(M_0, a, b, P_0=1, c, d, n=1:8000)
{
if (a == c)
   print(paste0("Solution not valid when a == c ==", a))
l1=b/(1-a)*d/(1-c)
l2=b*(P_0*(1-c)-d)/(c*(1-c))
M_n=(M_0-l2-l1)*a^n+l2*c^(n-1)+l1

return(M_n)
}


final_loc=function(M_0, a, b, P_0=1, c, d)
{
l1=b/(1-a)*d/(1-c)

return(l1)
}


plot(LOC(1e4, 0.95, 1e3),
	xlab="N", ylab="LOC")

t=dev_loc(1e4, 0.95, 1e3, 1, 0.95, 1)
plot(dev_loc(1e4, 0.98, 1e3, 1, 0.95, 1),
	xlab="N", ylab="LOC")

plot(dev_loc(1e2, 0.81, 1e2, 1, 0.37, 0.2, n=1:100),
        xlab="N", ylab="LOC")
plot(dev_loc(1e1, 0.79, 2e2, 1, 0.97, 0.1, n=1:100),
        xlab="N", ylab="LOC")
plot(dev_loc(1e1, 0.14, 8800, 1, 0.786, 0.203, n=1:100),
        xlab="N", ylab="LOC")

ks=read.csv("kernel-LOC.csv")

ks$released=as.Date(ks$released, format="%d/%m/%y")

plot(ks$released, ks$lines.added, col=pal_col[1], log="y",
	ylim=c(5e4, 1.7e6),
	xlab="Release date", ylab="LOC")
points(ks$released, ks$lines.removed, col=pal_col[2])

plot(ks$released, ks$developers, col=pal_col[1], log="y",
	ylim=c(50, 2100),
	xlab="Release date", ylab="People")
points(ks$released, ks$employers, col=pal_col[2])


# al_mod=glm(log(lines.added) ~ developers+released, data=ks)

al_mod=glm(lines.added ~ developers+released+log(lines), data=ks)
summary(al_mod)

ks$prev_lines=c(NA, tail(ks$lines, -1))
dl_mod=glm(log(lines.removed) ~ developers+released+log(prev_lines), data=ks)
summary(dl_mod)


# glibc

# id,branch_id,date,loc,sloc,files
pal_col=rainbow(2)

glibc=read.csv("glibc_evo.csv.xz", as.is=TRUE)

glibc$date=as.Date(glibc$date, format="%Y-%m-%d")
start_date=as.Date("1990-01-01", format="%Y-%m-%d")
glibc$Number_days=as.integer(difftime(glibc$date,
                                         start_date,
                                         units="days"))

glibc_main=subset(glibc, branch_id == 1)
glibc_main=subset(glibc_main, Number_days > 0)
x_bounds=min(glibc_main$Number_days):max(glibc_main$Number_days)

# They are probably already sorted by date
glibc_main=glibc_main[order(glibc_main$Number_days), ]

glibc_main$ksloc=glibc_main$sloc/1e3

plot(glibc_main$Number_days, glibc_main$ksloc, col=pal_col[2],
        xaxs="i", yaxs="i",
        xlab="Days since 1 Jan 1990", ylab="KSLOC")

l_mod=nls(ksloc ~ SSfpl(Number_days, a, b, c, d), data=glibc_main)
pred=predict(l_mod, list(Number_days=x_bounds))

lines(x_bounds, pred, col=pal_col[1])

dev.copy(dev=png, file="code-growth-glibc.png")
dev.off()


# l1=b/(1-a)*d/(1-c)
# l2=b*(P_0*(1-c)-d)/(c*(1-c))
# M_n=(M_0-l2-l1)*a^n+l2*c^(n-1)+l1
glibcQtrs=subset(glibc_main, Number_days %% 90 == 0)
glibcQtrs$Number_qtrs=glibcQtrs$Number_days %/% 90

# Didn't get this to fit.
pad_mod=nls(ksloc ~ (10-b*(1*(1-c)-d)/(c*(1-c))-b/(1-a)*d/(1-c))*a^Number_qtrs+b*(1*(1-c)-d)/(c*(1-c))*c^(Number_qtrs-1)+b/(1-a)*d/(1-c),
			data=glibcQtrs, trace=TRUE,
			start=list(a=0.79, b = 100, c=0.97, d=0.1))
summary(pad_mod)


